/**
 * Copyright (c) 2009 Barco n.v. All Rights Reserved. This software is confidential and proprietary information of Barco n.v. ("Confidential
 * Information"). You shall not disclose such Confidential Information and shall use it only in accordance with the terms of the license agreement you
 * entered into with Barco.
 */
package test;


import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.WeakHashMap;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.Topic;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;
import org.jnp.interfaces.NamingContext;



/**
 * This Util class used to fire internal/external events.
 * 
 * @author jmra,VESS
 */
public class EventEmitter
{
    private static final Logger mLogger = Logger.getLogger(EventEmitter.class);
    /**
     * TODO: Introduced for testing will be removed after event order understood
     */
    private static int jmsSessionId = 0;

    /**
     * Connection factory to get Queue/Topic
     */
    private static final String _TX_CONNECTION_FACTORY = "java:/JmsXA";
    private static final String _NONTX_CONNECTION_FACTORY = "ConnectionFactory";

    /**
     * CMS Public Topic to push/listen cms external events
     */
    public static final String _DESTINATION_CMS_PUBLIC_ = "topic/CMSPublic";

    /**
     * CMS Management Queue to push/listen cms internal events
     */
    public static final String _DESTINATION_CMS_MANAGEMENT_ = "queue/CMSManagement";

//    private static ICmsCacheServiceMBean mCmsCacheService;

    /**
     * This is introuduced to solve event order issue, this map will collect all events from current transaction key as current thread. and all the
     * collected events will be flushed at the end of the transaction.
     */
    public static final Map<Thread, List<String>> currentThreadEventsMap = new WeakHashMap<Thread, List<String>>();


    /**
     * publish to client topic
     * 
     * @param anEvent
     */
    public static void publishClientEvent (final String anEvent)
    {
        try
        {
//            EventName eventName = anEvent.getEventName();
//            boolean isRestoreEvent = (eventName == EventName.RestoreProgress || eventName == EventName.RestoreStarted || eventName == EventName.DatabaseResetStarted);
//            boolean isServerEvent = (eventName == EventName.JbossStarted || eventName == EventName.JbossStopped);
//            if (isPublishing() || isRestoreEvent || isServerEvent || (eventName == EventName.LicenseExpired))
//            {
                // anEvent.setTransactional(false);
                sendJMSEvent(_DESTINATION_CMS_PUBLIC_, anEvent);
                mLogger.debug("Published event to Client "+ anEvent);
//            }
        }
        catch (JMSException e)
        {
            handleJMSException(anEvent, e);
        }
    }


    /**
     * To Publish and flush events
     * 
     * @param anEvent
     */
    public static void publishEventNow (final String anEvent)
    {
        publishEvent(anEvent);
        publishClientEvent(anEvent);
        clearEvent();
    }


    /**
     * Static method for DOM to fire events on JMS event queue.
     * 
     * @param anEvent Internal/Server event
     */
    public static void publishEvent (final String anEvent)
    {
        if (!isPublishing())
        {
            mLogger.trace("Publish event returned, Restore in progress : "+ anEvent);
            return;
        }

        try
        {
//            if (anEvent.getObjectId() != 0)
//            {
                Thread currentThread = Thread.currentThread();
                List<String> eventList = currentThreadEventsMap.get(currentThread);
                if (eventList == null)
                {
                    
                    eventList = new ArrayList<String>();
                }

//                if (!(anEvent instanceof SessionEvent) && anEvent.isTransactional())
//                {
//                    if (!anEvent.isInternal())
                    eventList.add(anEvent);
//                }

//                if (!anEvent.isInternal())
//                {
                    currentThreadEventsMap.put(currentThread, eventList);

//                }

                sendJMSEvent(_DESTINATION_CMS_MANAGEMENT_, anEvent);
//            }
        }
        catch (JMSException e)
        {
            handleJMSException(anEvent, e);
        }
    }


    /**
     * To stop internal events to the client.
     */
    public static void stopPublishing ()
    {
//        setPublishing(false);
        mLogger.debug("> Publishing events is stopped. Referenece value 'Publisher' is:"+ isPublishing());
    }


    /**
     * To Start client publishing
     */
    public static void startPublishing ()
    {
//        setPublishing(true);
        mLogger.debug("< Publishing events is started. Referenece value 'Publisher' is:"+ isPublishing());
    }


    /**
     * Return false if Restore operation in progress.
     */
    public static boolean isPublishing ()
    {
       return true;
    }


    // Helpers
    // =============================================================================================

    /**
     * Send multiple events in a session
     * 
     * @param aJMSDestinationString
     * @param anEventList
     * @param isTransactional
     * @throws JMSException
     */
    private static void sendJMSEvent (final String aJMSDestinationString, List<String> anEventList, final boolean isTransactional)
        throws JMSException
    {
        assert (!anEventList.isEmpty()) : "Event list cannot be empty.";

        Connection connection = null;
        Session session = null;
        MessageProducer producer = null;

        try
        {
            connection = getJMSConnection(isTransactional);

            final int transactedSession = (isTransactional) ? Session.SESSION_TRANSACTED : Session.AUTO_ACKNOWLEDGE;
            session = connection.createSession(isTransactional, transactedSession);
            Destination aJMSDestination = (Destination) getInitialContext().lookup(aJMSDestinationString);
            producer = session.createProducer(aJMSDestination);

            String jmsTopicSessionId = "";
            // TimeToLive is set on Topic to Transaction timeout period
            if (aJMSDestination instanceof Topic)
            {
                producer.setTimeToLive(System.currentTimeMillis() + 5 * 60 * 1000l);
                jmsTopicSessionId = "Group [" + getJMSSessionId() + "]-";
            }

            connection.start();

            int sessionEventId = 1;
            for (String event : anEventList)
            {
                String id = jmsTopicSessionId + "Event " + sessionEventId;
//                event.setId(id);
                producer.send(session.createObjectMessage(event+"\t"+sessionEventId));
                sessionEventId++;
            }
        }
        catch (NamingException e)
        {
            handleNamingException(anEventList, e);
        }
        finally
        {
            if (producer != null)
                try
                {
                    producer.close();
                }
                catch (Exception ignore)
                {
                }
            if (session != null)
                try
                {
                    session.close();
                }
                catch (Exception ignore)
                {
                }
            if (connection != null)
                try
                {
                    connection.close();
                }
                catch (Exception ignore)
                {
                }
        }
    }


    /**
     * Send single event per sessionr
     * 
     * @param aJMSDestinationString
     * @param anEvent
     * @param isTransactional
     * @throws JMSException
     */
    private static void sendJMSEvent (final String aJMSDestinationString, String anEvent)
        throws JMSException
    {
        assert (anEvent != null) : "Event list cannot be empty.";

        Connection connection = null;
        Session session = null;
        MessageProducer producer = null;

        try
        {
            connection = getJMSConnection(true);

            final int transactedSession = (true) ? Session.SESSION_TRANSACTED : Session.AUTO_ACKNOWLEDGE;
            session = connection.createSession(true, transactedSession);

            Destination aJMSDestination = (Destination) getInitialContext().lookup(aJMSDestinationString);
            producer = session.createProducer(aJMSDestination);

            /**
             * TODO: Introduced for testing will be removed after event order understood
             */
            String jmsTopicSessionId = "";
            // TimeToLive is set on Topic to Transaction timeout period
            if (aJMSDestination instanceof Topic)
            {
                producer.setTimeToLive(System.currentTimeMillis() + 5 * 60 * 1000l);
                jmsTopicSessionId = "Group [" + getJMSSessionId() + "]-";
            }

            connection.start();

            /**
             * TODO: Introduced for testing will be removed after event order understood
             */
//            String id = jmsTopicSessionId + "Event " + Utils.getCallId();
//            anEvent.setId(id);

            producer.send(session.createObjectMessage(anEvent));
        }
        catch (NamingException e)
        {
            handleNamingException(anEvent, e);
        }
        finally
        {
            if (producer != null)
                try
                {
                    producer.close();
                }
                catch (Exception ignore)
                {
                }
            if (session != null)
                try
                {
                    session.close();
                }
                catch (Exception ignore)
                {
                }
            if (connection != null)
                try
                {
                    connection.close();
                }
                catch (Exception ignore)
                {
                }
        }
    }


    private static Connection getJMSConnection (boolean isTransactional)
        throws NamingException, JMSException
    {
        ConnectionFactory connFactory = null;
        if (isTransactional)
        {
            connFactory = (ConnectionFactory) getInitialContext().lookup(_TX_CONNECTION_FACTORY);
        }
        else
        {
            connFactory = (ConnectionFactory) getInitialContext().lookup(_NONTX_CONNECTION_FACTORY);
        }

        return connFactory.createConnection();
    }


//    private static void setPublishing (boolean isCanPublish)
//    {
//        tryr
//        {
//            getCmsCacheSrvice().put(ICmsCacheServiceMBean.cPublishEventsFQN, ICmsCacheServiceMBean.cPublishEventsKey, new Boolean(isCanPublish));
//            if (!isCanPublish)
//            {
//                getCmsCacheSrvice().put(ICmsCacheServiceMBean.cRestoreStartedOnNodeFQN, ICmsCacheServiceMBean.cRestoreStartedOnNodeKey,
//                        getIPAddress());
//            }
//        }
//        catch (UnknownHostException anException)
//        {
//            mLogger.error("Exception occured while getting the System IP: ", anException);
//        }
//        catch (CacheException anException)
//        {
//            handleLookupException(CmsServiceLocator.CMS_CacheService, anException);
//        }
//        catch (Exception anException)
//        {
//            mLogger.error("Unknown error occured while updating value in the Cms Cache Service with FQN: ", ICmsCacheServiceMBean.cPublishEventsFQN,
//                    anException);
//            mCmsCacheService = null;
//        }
//    }


//    private static ICmsCacheServiceMBean getCmsCacrheSrvice ()
//    {
//        try
//        {
//            if (mCmsCacheService == null)
//            {
//                mCmsCacheService = (ICmsCacheServiceMBean) CmsServiceLocator.lookupService(CmsServiceLocator.CMS_CacheService,
//                        ICmsCacheServiceMBean.class);
//            }
//        }
//        catch (NamingException anException)
//        {
//            handleLookupException(CmsServiceLocator.CMS_CacheService, anException);
//        }
//        catch (MalformedObjectNameException anException)
//        {
//            handleLookupException(CmsServiceLocator.CMS_CacheService, anException);
//        }
//
//        return mCmsCacheService;
//    }


    /**
     * Gets the IPAddress for the current host
     * 
     * @return The IP address of the host
     * @throws UnknownHostException
     */
    private static String getIPAddress ()
        throws UnknownHostException
    {
        String systemName = System.getProperty("bind.address");
        mLogger.debug("Bind Address: "+ systemName);

        return InetAddress.getByName(systemName).getHostAddress();
    }


    /**
     * Exception handler while service lookup
     */
    private static void handleLookupException (String aServiceName, Exception anException)
    {
        mLogger.error("Error occured while lookup Cms Cache Service name: "+ aServiceName, anException);
    }


    private static void handleJMSException (Object aMsg, JMSException anException)
    {
        mLogger.error("JMS Exception occured JMS Event cant be published: "+ aMsg, anException);

        // throw new MessageFailure(aMsg, anException);
    }


    private static void handleNamingException (Object aMsg, NamingException anException)
    {
        mLogger.error("Naming Exception occured Event cantr be published ", anException);

        // throw new MessageFailure(aMsg, anException);
    }


    /**
     * To flush events at the end of parent transaction
     */
    public static void flushEvents ()
    {
        Thread curT = Thread.currentThread();
        List<String> eventList = currentThreadEventsMap.get(curT);
        if (eventList == null || eventList.isEmpty())
            return;

        try
        {
            if (isPublishing())
            {
                sendJMSEvent(_DESTINATION_CMS_PUBLIC_, eventList, true);
            }
        }
        catch (JMSException e)
        {
            handleJMSException(eventList, e);
        }
        finally
        {
            currentThreadEventsMap.remove(curT);
        }
    }


    /**
     * Clear the current thread context from the hashmap
     */
    public static void clearEvent ()
    {
        currentThreadEventsMap.remove(Thread.currentThread());
    }


    private static int getJMSSessionId ()
    {
        return jmsSessionId++;
    }
    
    private static InitialContext getInitialContext ()
            throws NamingException
        {
            final Properties props = new Properties();
            props.put(Context.INITIAL_CONTEXT_FACTORY, "org.jnp.interfaces.NamingContextFactory");
            props.put(Context.URL_PKG_PREFIXES, "org.jboss.naming:org.jnp.interfaces");
            
            /*
             * Socket timeout is set to 2 sec. as to reduce the chances of getting hung during the lookup if server is not available.
             */
            props.put("jnp.sotimeout", "2000");
//            props.put("jnp.timeout", "2000");
            props.put(Context.PROVIDER_URL, "192.168.56.14:1100,192.168.56.15:1100");
            props.put(NamingContext.JNP_DISABLE_DISCOVERY, "true");

            return new InitialContext(props);
        }

}
